@@ -61,6 +61,9 @@ class Compiler:
             # hook-cmd "command pattern" {function: {parms}, function: {parms}}
             ".*?hook-cmd \s*(\S.*) (\S.*) (\S.*)": self.hook_generator,
 
+            # hook-cmd-nr (non-recursive) "command pattern" {function: {parms}, function: {parms}}
+            ".*?hook-cmd-nr \s*(\S.*) (\S.*) (\S.*)": self.hook_nr_generator,
+
            # remove-hooks
             ".*?remove-hooks\s.*?(\S.*)?$": self.remove_hooks_generator,
 
@@ -99,17 +102,29 @@ class Compiler:
                 self.last_stmt = self.current_statement if self.current_statement.lstrip()[
                                                                0] not in nothingness else self.last_stmt
 
-
-    # Generate hook-cmd command
+    # Generate command hook
     def hook_generator(self, parms):
-        self.output.append(f'{parms["indentation"]}{watiba_ref}.spawn_ctlr.add_hook({parms["match"].group(1)}, {parms["match"].group(2)}, {parms["match"].group(3)})')
+        self.output.append(f'{parms["indentation"]}{watiba_ref}.add_hook(\
+            {parms["match"].group(1)}, \
+            {parms["match"].group(2)}, \
+            {parms["match"].group(3)}, \
+            True)')
     
-    # Remove command hooks
+    # Generate command non-recursive hook
+    def hook_nr_generator(self, parms):
+        self.output.append(f'{parms["indentation"]}{watiba_ref}.add_hook(\
+            {parms["match"].group(1)}, \
+            {parms["match"].group(2)}, \
+            {parms["match"].group(3)}, \
+            False)')
+    
+    # Generate removal of command hooks
     def remove_hooks_generator(self, parms):
         if len(parms["match"].groups()) >  0 and parms["match"].group(1) != None:
-            self.output.append(f'{parms["indentation"]}{watiba_ref}.spawn_ctlr.remove_hooks(parms["match"].group(1))')
-        else:
-            self.output.append(f'{parms["indentation"]}{watiba_ref}.spawn_ctlr.remove_hooks()')
+            self.output.append(f'{parms["indentation"]}{watiba_ref}.remove_hooks(parms["match"].group(1))')
+            return
+        
+        self.output.append(f'{parms["indentation"]}{watiba_ref}.remove_hooks()')
 
     # Generate chain command
     def chain_generator(self, parms):
