@@ -34,6 +34,9 @@ class Watiba(Exception):
     def __init__(self):
         self.spawn_ctlr = WTSpawnController()
         self.parms = {"ssh-port":22}
+        self.hooks = {}
+        self.hook_flags = {}
+        self.hook_mode = False
 
     def set_parms(self, args):
         for k,v in args.items():
@@ -42,23 +45,24 @@ class Watiba(Exception):
     # Called by spawned thread
     # Dir context is not kept by the spawn expression
     # Returns WTOutput object
-    def execute(self, cmd, host="localhost"):
+    def execute(self, command, host="localhost"):
         context = False
         if host == "localhost":
-            return self.bash(cmd, context)
+            return self.bash(command, context)
         else:
-            return self.ssh(cmd, host)
+            # A simple wrapper for self.bash()
+            return self.ssh(command, host)
 
     # Run command remotely
     # Returns WTOutput object
-    def ssh(self, cmd, host, context=True):
-        return self.bash(f'ssh -p {self.parms["ssh-port"]} {host} "{cmd}"', context)
+    def ssh(self, command, host, context=True):
+        return self.bash(f'ssh -p {self.parms["ssh-port"]} {host} "{command}"', context)
 
-    # cmd - command string to execute
+    # command - command string to execute
     # context = track or not track current dir
     # Returns:
     #   WTOutput object that encapsulates stdout, stderr, exit code, etc.
-    def bash(self, cmd, context=True):
+    def bash(self, command, context=True):
 
         # In order to be thread-safe in the generated code, ALWAYS create a new output object for each command
         #  This is because in the generated code, the object reference, "_watiba_", is global and needs to be in scope
@@ -66,9 +70,19 @@ class Watiba(Exception):
         # singleton with no side effects.
         out = WTOutput()
 
+        # Run any command hooks defined for this command
+        results = self.run_hooks(command)
+
+        # Handle any hook failures
+        # All hooks are always run, but any one reporting a failure will cause the command to not be run
+        if results['success'] != True:
+            msg = f"One or more hooks failed. Hooks reporting a problem: {', '.join(results['failed-hooks'])}"
+            out.stderr.append(msg)
+            raise Exception(msg)
+
         # Tack on this command to see what the current dir is after the user's command is executed
         ctx = ' && echo "_watiba_cwd_($(pwd))_"' if context else ''
-        p = Popen(f"{cmd}{ctx}",
+        p = Popen(f"{command}{ctx}",
                   shell=True,
                   stdout=PIPE,
                   stderr=PIPE,
@@ -102,59 +116,12 @@ class Watiba(Exception):
             # Link this child promise to its parent
             l_promise.relate(parent_locals['promise'])
         
-        # Run all the command pre-execution hooks
-        # If any hooks returns False, meaning it somehow failed (that's determined by the hook)
-        # then report so an exception is thrown by the caller
-        def run_hooks(hooks, command):
-
-            # object returned to caller:
-            #  "success" - Aggregate True/False of all hooks.  (i.e. any hook that reports False will cause this value to return False)
-            #  "failed-hooks" - Array of hook names that reported a False condition
-            #       Note: all hooks are called in order even after one reports failure (i.e. reports False)
-            return_obj = {"success": True, "failed-hooks": []}
-
-            # Loop through the hooks and run them.  Also track ones that fail (i.e. report a False return code)
-            for cmd_regex, functions in hooks.items():
-
-                # Check if the command passed to us matches the regex expression
-                mat = re.match(cmd_regex, command)
-
-                # Does this command have attached hooks?
-                if mat:
-
-                    # Yes, command has hooks.  Run them.
-                    # If the hook fails track it, but keep going with the other hooks
-                    for func, parms in functions.items():
-
-                        # Call the hook.  The hook must return True if succeeded, False if failed
-                        rc = func(mat, parms)
-
-                        # If caller's hook didn't return a bool value, then it is marked as failed
-                        rc = False if type(rc) != bool else rc
-
-                        # Track failed hooks
-                        if rc == False:
-                            return_obj["failed-hooks"].append(func.__name__)
-                    
-                        return_obj["success"] &= rc
-            
-            return return_obj
-
 
         # This is run under the new thread, and under the control of wtspawncontroller.py (i.e. spawn controller calls this function)
         def run_command(promise, thread_args):
             # Get our thread id
             promise.thread_id = threading.get_ident()
 
-            # Run hooks, if any were defined
-            if len(thread_args["spawn-args"]["hooks"]) > 0:
-                ro = run_hooks(thread_args["spawn-args"]["hooks"], thread_args["command"])
-
-                # If any hook failed, throw an exception to our caller and name the bad hooks
-                if ro["success"] != True:
-                    raise Exception(f"These hooks failed: {', '.join(ro['failed-hooks'])}")
-
-
             # Execute the command in a new thread (this is synchronously run)
             promise.output = self.execute(thread_args["command"], thread_args["host"])
 
@@ -174,31 +141,70 @@ class Watiba(Exception):
 
         return l_promise
     
+    # Run all the command pre-execution hooks
+    # If any hooks returns False, meaning it somehow failed (that's determined by the hook)
+    # then report so an exception is thrown by the caller
+    def run_hooks(self, command):
 
+        # object returned to caller:
+        #  "success" - Aggregate True/False of all hooks.  (i.e. any hook that reports False will cause this value to return False)
+        #  "failed-hooks" - Array of hook names that reported a False condition
+        #       Note: all hooks are called in order even after one reports failure (i.e. reports False)
+        return_obj = {"success": True, "failed-hooks": []}
 
-    # Pipe either stdout or stderr to some target host with some target cmd
+        # Loop through the hooks and run them.  Also track ones that fail (i.e. report a False return code)
+        for command_regex, functions in self.hooks.items():
+
+            if self.hook_flags[command_regex]["recursive"] == False and self.hook_mode == True:
+                continue
+
+            # Check if the command passed to us matches the regex expression
+            mat = re.match(command_regex, command)
+
+            # Does this command have attached hooks?
+            if mat:
+
+                # Yes, command has hooks.  Run them.
+                # If the hook fails track it, but keep going with the other hooks
+                for func, parms in functions.items():
+
+                    # Call the hook.  The hook must return True if succeeded, False if failed
+                    rc = func(mat, parms)
+
+                    # If caller's hook didn't return a bool value, then it is marked as failed
+                    rc = False if type(rc) != bool else rc
+
+                    # Track failed hooks
+                    if rc == False:
+                        return_obj["failed-hooks"].append(func.__name__)
+                
+                    return_obj["success"] &= rc
+        
+        return return_obj
+
+    # Pipe either stdout or stderr to some target host with some target command
     def pipe(self, pipe_source, pipe_target):
         # Pipe output to target host command
-        for pipe_to, cmd in pipe_target.items():
+        for pipe_to, command in pipe_target.items():
             for line in pipe_source:
                 # The output for piped command is not kept, but is checked for the exit code
-                out = self.ssh(f'echo "{line}" | {cmd}', pipe_to)
+                out = self.ssh(f'echo "{line}" | {command}', pipe_to)
                 if out.exit_code != 0:
                     raise WTChainException(f'Piped command failed on {pipe_to}.  Error code: {out.exit_code}', pipe_to,
-                                           cmd, out)
+                                           command, out)
 
     # chain commands across various servers.  (Run sequentially and with regard to exit code.  A bad exit code causes
     # an exception to be thrown.
     #  A dictionary structure must be passed by the user's program as follows:
     #       {"hosts": ["host1", "host2", ...],  # These are the hosts to run the command on and is required
-    #        "stdout": {"source-host": {"target-host1":cmd, "target-host2":cmd, ...}}, # Pipe stdout from source to target(s) (optional)
-    #        "stderr": {"source-host": {"target-host1":cmd, "target-host2":cmd, ...}}   # Pipe stderr from source to target(s) (optional)
+    #        "stdout": {"source-host": {"target-host1":command, "target-host2":command, ...}}, # Pipe stdout from source to target(s) (optional)
+    #        "stderr": {"source-host": {"target-host1":command, "target-host2":command, ...}}   # Pipe stderr from source to target(s) (optional)
     #       }
     # Returns dictionary of WTOutput objects by host name: {host:WTOutput, ...}
-    def chain(self, cmd, parms):
+    def chain(self, command, parms):
         output = {}
         if "hosts" not in parms:
-            raise WTChainException("No hosts in argument dict", "none", cmd, None)
+            raise WTChainException("No hosts in argument dict", "none", command, None)
 
         pipe_stdout = parms["stdout"] if "stdout" in parms else {}
         pipe_stderr = parms["stderr"] if "stderr" in parms else {}
@@ -206,11 +212,11 @@ class Watiba(Exception):
         # Loop through each host and run the command on it
         for host in parms["hosts"]:
             # Run command remotely through SSH
-            output[host] = self.ssh(cmd, host)
+            output[host] = self.ssh(command, host)
 
             # If the command fails, bomb the whole execution
             if output[host].exit_code != 0:
-                raise WTChainException(f'Command failed on {host}. Error code: {output[host].exit_code}', host, cmd,
+                raise WTChainException(f'Command failed on {host}. Error code: {output[host].exit_code}', host, command,
                                        output[host])
 
             # If we are supposed to pipe the stdout for this host, do it
@@ -222,3 +228,41 @@ class Watiba(Exception):
                 self.pipe(output[host].stderr, pipe_stderr)
 
         return output
+
+
+    # Add a new hook.  If command pattern already exists, add the functions to it otherwise create a new pattern level.
+    #  Set recursive to True to keep hook from looping because it has a matching command within it
+    def add_hook(self, pattern, function, parms, recursive = True):
+
+        defined_pattern = pattern in self.hooks
+        defined_function = function in self.hooks[pattern] if defined_pattern else False
+
+        # If the command pattern and hook function are new, then add them
+        if defined_pattern and not defined_function:
+            self.hooks[pattern].update({function: parms})
+            return
+        
+        # If the command pattern already has this hook function 
+        # defined, just update the parms for the hook function
+        #  Note: defined_function cannot be True if defined_pattern is false
+        if defined_function:
+            self.hooks[pattern][function] = parms
+            return
+        
+        # At this point we know the pattern has not been defined yet, so define it
+        self.hooks.update({pattern : {function: parms}})
+        self.hook_flags[pattern] = {"recursive": recursive}
+
+    
+    # Remove a specific hook, keyed by pattern, or all hooks if no pattern is passed
+    def remove_hooks(self, pattern = None):
+        if not pattern:
+            self.hooks = {}
+            self.hook_flags = {}
+            return
+
+        if pattern in self.hooks:
+            del self.hooks[pattern]
+            del self.hook_flags[pattern]
+            return
+        
\ No newline at end of file
